﻿using System;
using System.Diagnostics;
using System.Threading;

namespace net.praqma.troja.library
{
    public class Command
    {
        private static Filter filter;
        private static Log log;

        private string[] user;
        private string command;
        private string arguments;
        private EnvironmentVariable[] environment;
        private bool mergeOutput = false;
        private TimeSpan timeout;
        private string remark;
        private Process process;
        private DateTime startTime;
        private bool hasStarted;
        private bool wasKilled;
        private bool wasTimedOut;
        private string standardOutput;
        private string standardError;
        private Thread OutputMonitor;
        private Thread ErrorMonitor;
        private Thread TimeoutMonitor;
        private Guid id;

        public Command(string[] user, string command, string arguments, EnvironmentVariable[] environment, bool mergeOutput, TimeSpan timeout, string remark)
        {
            this.user = user;
            this.hasStarted = false;
            this.command = command;
            this.arguments = arguments;
            this.environment = environment;
            this.mergeOutput = mergeOutput;
            this.timeout = timeout;
            this.id = Guid.NewGuid();
            this.remark = remark;

            Authorize();
            Init();
            SetupEnvironment();
            log.Add(this);
        }

        public void Start(bool synchronous)
        {
            hasStarted = true;
            startTime = DateTime.Now;
            process.Start();

            OutputMonitor = new Thread(new ThreadStart(this.MonitorStandardOutput));
            ErrorMonitor = new Thread(new ThreadStart(this.MonitorStandardError));
            TimeoutMonitor = new Thread(new ThreadStart(this.MonitorTimeout));

            OutputMonitor.Priority = ThreadPriority.BelowNormal;
            ErrorMonitor.Priority = ThreadPriority.BelowNormal;
            TimeoutMonitor.Priority = ThreadPriority.BelowNormal;

            OutputMonitor.Start();
            ErrorMonitor.Start();
            TimeoutMonitor.Start();

            if (synchronous)
            {
                process.WaitForExit();
            }

            log.Update(this);
        }

        public bool Kill()
        {
            if (hasStarted && !process.HasExited)
            {
                process.Kill();
                bool dead = process.WaitForExit(1000);
                wasKilled = !WasTimedOut;
                return dead;
            }
            else
            {
                return true;
            }
        }

        public string GetLog()
        {
            return log.GetLog(this);
        }

        #region Private Methods

        private void Authorize()
        {
            if (filter == null)
            {
                throw new IllegalCommandException(Filter.FILTERERROR << 1);
            }

            int authCode = filter.ValidateCommand(user, command, arguments, environment);

            if (authCode != 0)
            {
                throw new IllegalCommandException(authCode);
            }
        }

        private void Init()
        {
            process = new Process();
            process.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
            process.StartInfo.CreateNoWindow = true;
            process.StartInfo.UseShellExecute = false;
            process.StartInfo.RedirectStandardOutput = true;
            process.StartInfo.RedirectStandardError = true;
            process.StartInfo.FileName = command;
            process.StartInfo.Arguments = mergeOutput ? arguments + " 2>&1" : arguments;

            process.Exited += new EventHandler(ProcessExited);
        }

        private void SetupEnvironment()
        {
            if (environment != null)
            {
                foreach (EnvironmentVariable variable in environment)
                {
                    System.Environment.SetEnvironmentVariable(variable.Name, variable.Value);
                }
            }
        }

        private void MonitorStandardOutput()
        {
            standardOutput = "";

            while (!process.StandardOutput.EndOfStream)
            {
                //Thread.Sleep(1000);
                standardOutput += process.StandardOutput.ReadLine() + "\n";
            }
        }

        private void MonitorStandardError()
        {
            standardError = "";

            while (!process.StandardError.EndOfStream)
            {
                //Thread.Sleep(1000);
                standardError += process.StandardError.ReadLine() + "\n";
            }
        }

        private void MonitorTimeout()
        {
            Thread.Sleep(timeout);

            if (!process.HasExited)
            {
                wasTimedOut = true;
                process.Kill();
                process.WaitForExit(1000);
            }
        }

        private void ProcessExited(object sender, EventArgs e)
        {
            if (TimeoutMonitor.IsAlive)
            {
                TimeoutMonitor.Abort();
            }

            OutputMonitor.Join(2000);
            ErrorMonitor.Join(2000);

            log.Update(this);
        }

        #endregion

        #region Static Properties

        public static Log Log
        {
            get { return log; }
        }

        public static string FilterPath
        {
            get
            {
                return (filter != null)
                    ? filter.FilePath
                    : null;
            }
            set
            {
                if (FilterPath != value)
                    filter = new Filter(value);
            }
        }



        public static string LogPath
        {
            get
            {
                return (log != null)
                    ? log.FilePath
                    : null;
            }
            set
            {
                if (LogPath != value)
                    log = new Log(value);
            }
        }

        #endregion

        #region Properties

        public string User
        {
            get { return user[0]; }
        }

        public Guid Id
        {
            get { return id; }
        }

        public string Executable
        {
            get { return command + " " + arguments; }
        }

        public EnvironmentVariable[] Environment
        {
            get { return environment; }
        }

        public bool MergeOutput
        {
            get { return mergeOutput; }
        }

        public TimeSpan Timeout
        {
            get { return timeout; }
        }

        public string Remark
        {
            get { return remark; }
        }

        public DateTime StartTime
        {
            get { return startTime; }
        }

        public DateTime ExitTime
        {
            get { return (hasStarted && process.HasExited) ? process.ExitTime : startTime; }
        }

        public bool HasStarted
        {
            get { return hasStarted; }
        }

        public bool HasExited
        {
            get { return hasStarted && process.HasExited; }
        }

        public string StandardOutput
        {
            get { return standardOutput; }
        }

        public string StandardError
        {
            get { return standardError; }
        }

        public TimeSpan Duration
        {
            get
            {
                return (!hasStarted)
                ? new TimeSpan(0, 0, 0)
                : (process.HasExited)
                    ? process.ExitTime.Subtract(startTime)
                    : DateTime.Now.Subtract(startTime);
            }
        }

        public int ExitCode
        {
            get
            {
                return (hasStarted && process.HasExited)
                    ? process.ExitCode
                    : 256;
            }
        }

        public bool WasKilled
        {
            get { return wasKilled; }
        }

        public bool WasTimedOut
        {
            get { return wasTimedOut; }
        }

        #endregion
    }
}
