﻿using System;
using System.Collections.Generic;
using System.Security.Principal;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Schema;
using net.praqma.systemlog;

namespace net.praqma.troja.library
{
    public class Filter
    {
        public static readonly int FILTERERROR = 1;
        public static readonly int DENIEDUSER = 2;
        public static readonly int DENIEDCOMMAND = 3;
        public static readonly int DENIEDENVIRONMENT = 4;
        public static readonly int UNKNOWNCOMMAND = 5;
        public static readonly int UNALLOWEDARGUMENT = 6;
        public static readonly int UNALLOWEDENVIRONMENT = 7;
        public static readonly int UNALLOWEDUSER = 8;

        private static readonly string XMLNS = @"http://praqma.net/troja/filter/";
        private static readonly string VERSION = "1.2";
        private static readonly string PREFIX = "praq";
        private static readonly string XSD = Configuration.FilterSchemaUrl;
        private static readonly char[] WHITESPACES = { ' ', '\t' };

        private Blacklist blacklist;
        private WhiteList whitelist;
        private string filePath;

        internal Filter(string filePath)
        {
            try
            {
                this.filePath = filePath;
                Unmarshall();
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(WindowsIdentity.GetCurrent().Name, ex);
                throw new IllegalCommandException(1 << FILTERERROR);
            }
        }

        public string FilePath
        {
            get { return filePath; }
        }

        private void Unmarshall()
        {
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.Schemas.Add(XMLNS, XSD);
            settings.ValidationEventHandler += new ValidationEventHandler(ValidationCallback);
            settings.ValidationType = ValidationType.Schema;

            XmlReader reader = XmlReader.Create(FilePath, settings);
            XmlNamespaceManager namespaceManager = new XmlNamespaceManager(reader.NameTable);
            namespaceManager.AddNamespace(PREFIX, XMLNS);

            XmlDocument document = new XmlDocument();
            document.Load(reader);
            reader.Close();

            XmlElement root = document.DocumentElement;

            UnmarshallBlacklist(root, namespaceManager);
            UnmarshallWhitelist(root, namespaceManager);
        }

        private void UnmarshallBlacklist(XmlElement root, XmlNamespaceManager namespaceManager)
        {
            blacklist = new Blacklist();

            XmlNode blacklistNode = root.SelectSingleNode(PREFIX + ":blacklist", namespaceManager);

            XmlNodeList commands = blacklistNode.SelectNodes(PREFIX + ":executables/" + PREFIX + ":pattern", namespaceManager);
            foreach (XmlNode node in commands)
            {
                blacklist.AddCommand(node.InnerText);
            }

            XmlNodeList keys = blacklistNode.SelectNodes(PREFIX + ":environment/" + PREFIX + ":key", namespaceManager);
            foreach (XmlNode node in keys)
            {
                blacklist.AddKey(node.InnerText);
            }

            XmlNodeList values = blacklistNode.SelectNodes(PREFIX + ":environment/" + PREFIX + ":value", namespaceManager);
            foreach (XmlNode node in values)
            {
                blacklist.AddValue(node.InnerText);
            }

            XmlNodeList users = blacklistNode.SelectNodes(PREFIX + ":users/" + PREFIX + ":entry[@type='user']", namespaceManager);
            foreach (XmlNode node in users)
            {
                blacklist.AddUser(node.InnerText);
            }

            XmlNodeList groups = blacklistNode.SelectNodes(PREFIX + ":users/" + PREFIX + ":entry[@type='group']", namespaceManager);
            foreach (XmlNode node in groups)
            {
                blacklist.AddGroup(node.InnerText);
            }
        }

        private void UnmarshallWhitelist(XmlElement root, XmlNamespaceManager namespaceManager)
        {
            whitelist = new WhiteList();

            //without arguments
            XmlNodeList commands = root.SelectNodes(PREFIX + ":whitelist/" + PREFIX + ":command[" + PREFIX + ":noArguments]", namespaceManager);

            foreach (XmlNode cmdNode in commands)
            {
                FilterCommand command = new FilterCommand(false);
                command.Executable = cmdNode.SelectSingleNode(PREFIX + ":executable", namespaceManager).InnerText;

                XmlNodeList keys = cmdNode.SelectNodes(PREFIX + ":environment/" + PREFIX + ":key", namespaceManager);
                foreach (XmlNode node in keys)
                {
                    command.AddKey(node.InnerText);
                }

                XmlNodeList values = cmdNode.SelectNodes(PREFIX + ":environment/" + PREFIX + ":value", namespaceManager);
                foreach (XmlNode node in values)
                {
                    command.AddValue(node.InnerText);
                }

                XmlNodeList users = cmdNode.SelectNodes(PREFIX + ":users/" + PREFIX + ":entry[@type='user']", namespaceManager);
                foreach (XmlNode node in users)
                {
                    command.AddUser(node.InnerText);
                }

                XmlNodeList groups = cmdNode.SelectNodes(PREFIX + ":users/" + PREFIX + ":entry[@type='group']", namespaceManager);
                foreach (XmlNode node in groups)
                {
                    command.AddGroup(node.InnerText);
                }

                whitelist.AddCommand(command);
            }

            //with arguments
            commands = root.SelectNodes(PREFIX + ":whitelist/" + PREFIX + ":command[" + PREFIX + ":arguments]", namespaceManager);

            foreach (XmlNode cmdNode in commands)
            {
                FilterCommand command = new FilterCommand(true);
                command.Executable = cmdNode.SelectSingleNode(PREFIX + ":executable", namespaceManager).InnerText;

                XmlNodeList arguments = cmdNode.SelectNodes(PREFIX + ":arguments/" + PREFIX + ":argument", namespaceManager);
                foreach (XmlNode node in arguments)
                {
                    command.AddArgument(node.InnerText);
                }

                XmlNodeList keys = cmdNode.SelectNodes(PREFIX + ":environment/" + PREFIX + ":key", namespaceManager);
                foreach (XmlNode node in keys)
                {
                    command.AddKey(node.InnerText);
                }

                XmlNodeList values = cmdNode.SelectNodes(PREFIX + ":environment/" + PREFIX + ":value", namespaceManager);
                foreach (XmlNode node in values)
                {
                    command.AddValue(node.InnerText);
                }

                XmlNodeList users = cmdNode.SelectNodes(PREFIX + ":users/" + PREFIX + ":entry[@type='user']", namespaceManager);
                foreach (XmlNode node in users)
                {
                    command.AddUser(node.InnerText);
                }

                XmlNodeList groups = cmdNode.SelectNodes(PREFIX + ":users/" + PREFIX + ":entry[@type='group']", namespaceManager);
                foreach (XmlNode node in groups)
                {
                    command.AddGroup(node.InnerText);
                }

                whitelist.AddCommand(command);
            }
        }

        internal int ValidateCommand(string[] user, string command, string arguments, EnvironmentVariable[] environment)
        {
            int errorcode = 0;

            if (blacklist.DenyUser(user))
            {
                errorcode += 1 << DENIEDUSER;
            }

            if (blacklist.DenyCommand(command, arguments))
            {
                errorcode += 1 << DENIEDCOMMAND;
            }

            if (blacklist.DenyEnvironment(environment))
            {
                errorcode += 1 << DENIEDENVIRONMENT;
            }

            if (errorcode != 0)
            {
                return errorcode;
            }

            LinkedList<FilterCommand> commands = whitelist.GetCommands(command);

            if (commands == null || commands.Count == 0)
            {
                return 1 << UNKNOWNCOMMAND;
            }
            else
            {
                bool okArguments = true;
                bool okEnvironment = true;
                bool okUser = true;

                foreach (FilterCommand cmd in commands)
                {
                    bool allowArguments = cmd.AllowArguments(arguments);
                    bool allowEnvironment = cmd.AllowEnvironment(environment);
                    bool allowUser = cmd.AllowUser(user);

                    if (allowArguments && allowEnvironment && allowUser)
                    {
                        return 0;
                    }
                    else
                    {
                        okArguments &= allowArguments;
                        okEnvironment &= allowEnvironment;
                        okUser &= allowUser;
                    }
                }

                if (!okArguments)
                {
                    errorcode += 1 << UNALLOWEDARGUMENT;
                }

                if (!okEnvironment)
                {
                    errorcode += 1 << UNALLOWEDENVIRONMENT;
                }

                if (!okUser)
                {
                    errorcode += 1 << UNALLOWEDUSER;
                }

                return errorcode;
            }
        }

        private void ValidationCallback(object sender, ValidationEventArgs args)
        {
            throw new XmlException(args.Message);
        }

        private class Blacklist
        {
            private LinkedList<Regex> commands;
            private LinkedList<Regex> keys;
            private LinkedList<Regex> values;
            private IDictionary<string, bool> users;
            private IDictionary<string, bool> groups;

            internal Blacklist()
            {
                this.commands = new LinkedList<Regex>();
                this.keys = new LinkedList<Regex>();
                this.values = new LinkedList<Regex>();
                this.users = new Dictionary<string, bool>();
                this.groups = new Dictionary<string, bool>();
            }

            internal void AddCommand(string command)
            {
                commands.AddLast(new Regex(command, RegexOptions.Compiled | RegexOptions.IgnoreCase));
            }

            internal void AddKey(string key)
            {
                keys.AddLast(new Regex(key, RegexOptions.Compiled | RegexOptions.IgnoreCase));
            }

            internal void AddValue(string value)
            {
                values.AddLast(new Regex(value, RegexOptions.Compiled | RegexOptions.IgnoreCase));
            }

            internal void AddUser(string user)
            {
                users.Add(user.ToLower(), true);
            }

            internal void AddGroup(string group)
            {
                groups.Add(group.ToLower(), true);
            }

            internal bool DenyCommand(string command, string arguments)
            {
                string cmdarg = (command + " " + arguments);

                foreach (Regex regex in commands)
                {
                    if (regex.IsMatch(cmdarg))
                    {
                        return true;
                    }
                }

                return false;
            }

            internal bool DenyEnvironment(EnvironmentVariable[] environment)
            {
                if (environment == null)
                {
                    return false;
                }

                foreach (EnvironmentVariable variable in environment)
                {
                    foreach (Regex regex in keys)
                    {
                        if (regex.IsMatch(variable.Name))
                        {
                            return true;
                        }
                    }

                    foreach (Regex regex in values)
                    {
                        if (regex.IsMatch(variable.Value))
                        {
                            return true;
                        }
                    }
                }

                return false;
            }

            internal bool DenyUser(string[] user)
            {
                if (users.ContainsKey(user[0].ToLower()))
                {
                    return true;
                }

                for (int i = 1; i < user.Length; i++)
                {
                    if (groups.ContainsKey(user[i].ToLower()))
                    {
                        return true;
                    }
                }

                return false;
            }
        }

        private class WhiteList
        {
            private Dictionary<string, LinkedList<FilterCommand>> commands;

            internal WhiteList()
            {
                commands = new Dictionary<string, LinkedList<FilterCommand>>();
            }

            internal void AddCommand(FilterCommand command)
            {
                string executable = command.Executable.ToLower();

                if (!commands.ContainsKey(executable))
                    commands[executable] = new LinkedList<FilterCommand>();

                commands[executable].AddLast(command);
            }

            internal LinkedList<FilterCommand> GetCommands(string executable)
            {
                executable = executable.ToLower();

                if (commands.ContainsKey(executable))
                    return commands[executable];

                return null;
            }
        }

        private class FilterCommand
        {
            private string executable;
            private LinkedList<Regex> arguments;
            private LinkedList<Regex> keys;
            private LinkedList<Regex> values;
            private IDictionary<string, bool> users;
            private IDictionary<string, bool> groups;

            internal FilterCommand(bool withArguments)
            {
                this.arguments = (withArguments)
                    ? new LinkedList<Regex>()
                    : null;
                this.keys = new LinkedList<Regex>();
                this.values = new LinkedList<Regex>();
                this.users = new Dictionary<string, bool>();
                this.groups = new Dictionary<string, bool>();
            }

            internal string Executable
            {
                get { return executable; }
                set { executable = value.ToLower(); }
            }

            internal void AddArgument(string argument)
            {
                arguments.AddLast(new Regex(argument, RegexOptions.Compiled | RegexOptions.IgnoreCase));
            }

            internal void AddKey(string key)
            {
                keys.AddLast(new Regex(key, RegexOptions.Compiled | RegexOptions.IgnoreCase));
            }

            internal void AddValue(string value)
            {
                values.AddLast(new Regex(value, RegexOptions.Compiled | RegexOptions.IgnoreCase));
            }

            internal void AddUser(string user)
            {
                users.Add(user.ToLower(), true);
            }

            internal void AddGroup(string group)
            {
                groups.Add(group.ToLower(), true);
            }

            internal bool AllowArguments(string argument)
            {
                if (argument == null)
                {
                    return arguments == null;
                }

                argument = argument.TrimStart(WHITESPACES);

                if (arguments == null)
                {
                    return argument.Length == 0;
                }
                else
                {
                    foreach (Regex regex in arguments)
                    {
                        if (regex.IsMatch(argument))
                        {
                            return true;
                        }
                    }

                    return false;
                }
            }

            internal bool AllowEnvironment(EnvironmentVariable[] environment)
            {
                if (environment == null)
                {
                    return true;
                }

                foreach (EnvironmentVariable variable in environment)
                {
                    foreach (Regex regex in keys)
                    {
                        if (regex.IsMatch(variable.Name))
                        {
                            return true;
                        }
                    }

                    foreach (Regex regex in values)
                    {
                        if (regex.IsMatch(variable.Value))
                        {
                            return true;
                        }
                    }
                }

                return false;
            }

            internal bool AllowUser(string[] user)
            {
                if (users.ContainsKey(user[0]))
                {
                    return true;
                }

                for (int i = 1; i < user.Length; i++)
                {
                    if (groups.ContainsKey(user[i]))
                    {
                        return true;
                    }
                }

                return false;
            }
        }
    }
}
