﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Principal;
using System.Xml;
using System.Xml.Schema;
using net.praqma.systemlog;

namespace net.praqma.troja.library
{
    public class Log
    {
        private static readonly string XMLNS = @"http://praqma.net/troja/log/";
        private static readonly string VERSION = "1.0";
        private static readonly string PREFIX = "praq";
        private static readonly string XSD = Configuration.LogSchemaUrl;

        private string filePath;
        private XmlNamespaceManager namespaceManager;
        private XmlDocument document;

        internal Log(string filePath)
        {
            try
            {
                this.filePath = filePath;
                Unmarshall();
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(WindowsIdentity.GetCurrent().Name, ex);
                throw ex;
            }
        }

        public string FilePath
        {
            get { return filePath; }
        }

        private void Unmarshall()
        {
            if (File.Exists(FilePath))
            {
                XmlReaderSettings settings = new XmlReaderSettings();
                settings.Schemas.Add(XMLNS, XSD);
                settings.ValidationEventHandler += new ValidationEventHandler(ValidationCallback);
                settings.ValidationType = ValidationType.Schema;

                XmlReader reader = XmlReader.Create(filePath, settings);
                namespaceManager = new XmlNamespaceManager(reader.NameTable);
                namespaceManager.AddNamespace(PREFIX, XMLNS);

                document = new XmlDocument();
                document.Load(reader);
                reader.Close();
            }
            else
            {
                document = createDocument();
                saveChanges();
            }
        }

        private void ValidationCallback(object sender, ValidationEventArgs args)
        {
            throw new XmlException(args.Message);
        }
        
        public void Add(Command c)
        {
            XmlElement command = document.CreateElement(PREFIX, "command", XMLNS);
            document.DocumentElement.AppendChild(command);

            XmlAttribute id = document.CreateAttribute("id");
            id.Value = c.Id.ToString();
            command.Attributes.Append(id);

            XmlElement created = document.CreateElement(PREFIX, "created", XMLNS);
            created.InnerText = DateTime.Now.ToString("s");
            command.AppendChild(created);

            XmlElement user = document.CreateElement(PREFIX, "user", XMLNS);
            user.InnerText = c.User;
            command.AppendChild(user);

            XmlElement executable = document.CreateElement(PREFIX, "executable", XMLNS);
            executable.InnerText = c.Executable;
            command.AppendChild(executable);

            if (c.Environment != null && c.Environment.Length > 0)
            {
                XmlElement environment = document.CreateElement(PREFIX, "environment", XMLNS);
                command.AppendChild(environment);

                foreach (EnvironmentVariable env in c.Environment)
                {
                    XmlElement variable = document.CreateElement(PREFIX, "variable", XMLNS);
                    variable.InnerText = env.Value;
                    XmlAttribute name = document.CreateAttribute("name");
                    name.Value = env.Name;
                    variable.Attributes.Append(name);
                    environment.AppendChild(variable);
                }
            }

            XmlElement timeOut = document.CreateElement(PREFIX, "timeOut", XMLNS);
            timeOut.InnerText = string.Format("PT{0}H{1}M{2}S", c.Timeout.Hours, c.Timeout.Minutes, c.Timeout.Seconds);
            command.AppendChild(timeOut);

            if (c.Remark != null)
            {
                XmlElement remark = document.CreateElement(PREFIX, "remark", XMLNS);
                remark.InnerText = c.Remark;
                command.AppendChild(remark);
            }

            if (c.MergeOutput)
                command.AppendChild(document.CreateElement(PREFIX, "merged", XMLNS));

            saveChanges();
        }

        public string GetLog(Command c)
        {
            Update(c);

            XmlElement command = document.SelectSingleNode("//" + PREFIX + ":command[@id='" + c.Id + "']", namespaceManager) as XmlElement;

            if (command == null)
                return null;
            
            XmlDocument doc = createDocument();
            doc.DocumentElement.AppendChild(doc.ImportNode(command, true));

            return doc.OuterXml;
        }

        public string GetLog(ICollection<Command> commands)
        {
            XmlDocument doc = createDocument();

            foreach (Command c in commands)
            {
                Update(c);

                XmlElement command = document.SelectSingleNode("//" + PREFIX + ":command[@id='" + c.Id + "']", namespaceManager) as XmlElement;

                if (command == null)
                    continue;

                doc.DocumentElement.AppendChild(doc.ImportNode(command, true));
            }

            return doc.OuterXml;
        }

        private XmlDocument createDocument()
        {
            XmlDocument doc = new XmlDocument();
            doc.AppendChild(doc.CreateXmlDeclaration("1.0", "utf-8", null));

            namespaceManager = new XmlNamespaceManager(doc.NameTable);
            namespaceManager.AddNamespace(PREFIX, XMLNS);

            XmlElement log = doc.CreateElement(PREFIX, "log", XMLNS);
            doc.AppendChild(log);

            XmlAttribute version = doc.CreateAttribute("version");
            version.Value = VERSION;
            log.Attributes.Append(version);

            return doc;
        }

        private void saveChanges()
        {
            XmlWriterSettings settings = new XmlWriterSettings();
            settings.CloseOutput = true;
            settings.Indent = true;
            settings.IndentChars = "\t";

            XmlWriter writer = XmlWriter.Create(filePath, settings);
            document.WriteTo(writer);
            writer.Close();
        }
        
        public void Update(Command c)
        {
            if (c.HasStarted)
            {
                XmlElement command = document.SelectSingleNode("//" + PREFIX + ":command[@id='" + c.Id + "']", namespaceManager) as XmlElement;

                if (command == null)
                    throw new NullReferenceException("Unknown command: " + c.Id);

                XmlElement startTime = command.SelectSingleNode(PREFIX + ":startTime", namespaceManager) as XmlElement;

                if (startTime == null)
                {
                    XmlElement standardOut = document.CreateElement(PREFIX, "standardOut", XMLNS);
                    XmlCDataSection stdout = document.CreateCDataSection(c.StandardOutput);
                    standardOut.AppendChild(stdout);
                    command.AppendChild(standardOut);

                    XmlElement errorOut = document.CreateElement(PREFIX, "errorOut", XMLNS);
                    XmlCDataSection errout = document.CreateCDataSection(c.StandardError);
                    errorOut.AppendChild(errout);
                    command.AppendChild(errorOut);

                    startTime = document.CreateElement(PREFIX, "startTime", XMLNS);
                    startTime.InnerText = c.StartTime.ToString("s");
                    command.AppendChild(startTime);
                }
                else
                {
                    XmlElement standardOut = command.SelectSingleNode(PREFIX + ":standardOut", namespaceManager) as XmlElement;
                    XmlCDataSection stdout = document.CreateCDataSection(c.StandardOutput);
                    standardOut.RemoveAll();
                    standardOut.AppendChild(stdout);

                    XmlElement errorOut = command.SelectSingleNode(PREFIX + ":errorOut", namespaceManager) as XmlElement;
                    XmlCDataSection errout = document.CreateCDataSection(c.StandardError);
                    errorOut.RemoveAll();
                    errorOut.AppendChild(errout);
                }

                if (c.HasExited)
                {
                    XmlElement endTime = command.SelectSingleNode(PREFIX + ":endTime", namespaceManager) as XmlElement;

                    if (endTime == null)
                    {
                        endTime = document.CreateElement(PREFIX, "endTime", XMLNS);
                        endTime.InnerText = c.ExitTime.ToString("s");
                        command.AppendChild(endTime);

                        XmlElement exitCode = document.CreateElement(PREFIX, "exitCode", XMLNS);
                        exitCode.InnerText = c.ExitCode.ToString();
                        command.AppendChild(exitCode);

                        if (c.WasKilled)
                            command.AppendChild(document.CreateElement(PREFIX, "killed", XMLNS));

                        if (c.WasTimedOut)
                            command.AppendChild(document.CreateElement(PREFIX, "timedOut", XMLNS));
                    }
                }

                saveChanges();
            }
        }

    }
}
