﻿using System;
using System.Security.Principal;
using System.ServiceModel;
using net.praqma.systemlog;
using net.praqma.troja.library;
using System.Collections.Generic;
using System.Threading;

namespace net.praqma.troja.wcfservice
{
    public class TrojaWcf : ITrojaWcf
    {
        private static IDictionary<string, IDictionary<Guid, Command>> userPool =
            new Dictionary<string, IDictionary<Guid, Command>>();
        private static Thread poolKeeper;
        
        #region Construction

        public TrojaWcf()
        {
            if ((poolKeeper == null) || !poolKeeper.IsAlive)
            {
                poolKeeper = new Thread(new ThreadStart(this.PoolCleanup));
                poolKeeper.Priority = ThreadPriority.Lowest;
                poolKeeper.Start();
            }
        }

        #endregion

        #region ITrojaWcf Members

        //public Guid CreateCommand(string command, string arguments, bool merge, TimeSpan timeout, string remark)
        public Guid CreateCommand(string command, string arguments, bool merge, int hour, int minute, int second, string remark)
        {
            try
            {
                Command.FilterPath = Configuration.FilterUrl;
                Command.LogPath = Configuration.LogFile;
                //Command cmd = new Command(getMemberships(), command, arguments, null, merge, timeout, remark);
                Command cmd = new Command(getMemberships(), command, arguments, null, merge, new TimeSpan(hour, minute, second), remark);
                UserCommands().Add(cmd.Id, cmd);

                return cmd.Id;
            }
            catch (IllegalCommandException ex)
            {
                SystemLog.WriteWarning(getUser(), ex);
                throw ex;
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(getUser(), ex);
                throw new TrojaException("Could not create the command: " + ex.GetType().Name + " " + ex.Message);
            }
        }

        public string GetCommands()
        {
            try
            {
                Command.LogPath = Configuration.LogFile;
                return Command.Log.GetLog(UserCommands().Values);
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(getUser(), ex);
                throw new TrojaException("Could not retrieve commands.");
            }
        }

        public void StartAsync(Guid id)
        {
            Command command = UserCommands()[id];

            if (command == null)
            {
                Exception ex = new TrojaException("Unknown command: " + id);
                SystemLog.WriteWarning(getUser(), ex);
                throw ex;
            }

            try
            {
                command.Start(false);
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(getUser(), ex);
                throw new TrojaException("Unable to start command.");
            }
        }

        public string StartSync(Guid id)
        {
            Command command = UserCommands()[id];

            if (command == null)
            {
                Exception ex = new TrojaException("Unknown command: " + id);
                SystemLog.WriteWarning(getUser(), ex);
                throw ex;
            }

            try
            {
                command.Start(true);
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(getUser(), ex);
                throw new TrojaException("Unable to start command.");
            }

            return command.GetLog();
        }

        public string GetLog(Guid id)
        {
            Command command = UserCommands()[id];

            if (command == null)
            {
                Exception ex = new TrojaException("Unknown command: " + id);
                SystemLog.WriteWarning(getUser(), ex);
                throw ex;
            }

            try
            {
                return command.GetLog();
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(getUser(), ex);
                throw new TrojaException("Could not retrieve log.");
            }
        }

        public string KillCommand(Guid id)
        {
            Command command = UserCommands()[id];

            if (command == null)
            {
                Exception ex = new TrojaException("Unknown command: " + id);
                SystemLog.WriteWarning(getUser(), ex);
                throw ex;
            }

            try
            {
                command.Kill();
            }
            catch (Exception ex)
            {
                SystemLog.WriteError(getUser(), ex);
                throw new TrojaException("Could not kill command: " + id);
            }

            return command.GetLog();
        }

        public string WhoAmI()
        {
            return getUser();
        }

        public string ServiceUser()
        {
            return WindowsIdentity.GetCurrent().Name;
        }

        #endregion

        #region Private methods

        private IDictionary<Guid, Command> UserCommands()
        {
            string user = getUser();

            IDictionary<Guid, Command> commands;

            if (userPool.ContainsKey(user))
            {
                commands = userPool[user];
            }
            else
            {
                commands = new Dictionary<Guid, Command>();
                userPool[user] = commands;
            }

            return commands;
        }

        private string getUser()
        {
            try
            {
                return ServiceSecurityContext.Current.WindowsIdentity.Name;
            }
            catch (Exception)
            {
                Exception ex = new TrojaException("Could not determine the caller WindowsIdentity");
                SystemLog.WriteError(getUser(), ex);
                throw ex;
            }
        }

        private string[] getMemberships()
        {
            try
            {
                List<string> groups = new List<string>();
                groups.Add(getUser());
                foreach (IdentityReference group in ServiceSecurityContext.Current.WindowsIdentity.Groups)
                {
                    groups.Add(group.Translate(typeof(NTAccount)).ToString().ToLower());
                }
                return groups.ToArray();
            }
            catch (Exception)
            {
                Exception ex = new TrojaException("Could not determine the caller WindowsIdentity");
                SystemLog.WriteError(getUser(), ex);
                throw ex;
            }
        }

        /*
         * Cleans out dead commands once every hour.
         */
        private void PoolCleanup()
        {
            while (true)
            {
                foreach (KeyValuePair<string, IDictionary<Guid, Command>> user in userPool)
                {
                    foreach (KeyValuePair<Guid, Command> command in user.Value)
                    {
                        Command cmd = command.Value;

                        if ((cmd.HasExited) && (DateTime.Now.Subtract(cmd.ExitTime).Hours >= 1))
                        {
                            user.Value.Remove(command.Key);
                        }
                    }

                    if (user.Value.Count == 0)
                    {
                        userPool.Remove(user.Key);
                    }
                }

                DateTime nextHour = DateTime.Today.AddHours(DateTime.Now.Hour + 1);
                Thread.Sleep(nextHour - DateTime.Now);
            }
        }

        #endregion
    }
}
